@@ -15,7 +15,7 @@ import types
 import warnings
 from weakref import WeakSet
 
-from htools import hdir, load, save, identity, hasstatic
+from htools import hdir, load, save, identity, hasstatic, tolist, select
 
 
 class AutoInit:
@@ -1389,7 +1389,7 @@ def valuecheck(func):
     return wrapper
 
 
-def debug(func=None, prefix='', arguments=True):
+def debug(func=None, prefix='', arguments=True, out_path=None):
     """Decorator that prints information about a function call. Often, this
     will only be used temporarily when debugging. Note that a wrapped function
     that accepts *args will display a signature including an 'args' parameter
@@ -1410,6 +1410,14 @@ def debug(func=None, prefix='', arguments=True):
     arguments: bool
         If True, the printed message will include the function arguments.
         If False, it will print the function name but not its arguments.
+    out_path: str or Path or None
+        If provided, a dict of arguments will be saved as a json file as
+        specified by this path. Intermediate directories will be created if
+        necessary. Function arguments will be made available for string
+        formatting if you wish to use that in the file name.
+        Example: 'data/models/{prefix}/args.json'. The argument "prefix" will
+        be used to save the file in the appropriate place. Note: `arguments`
+        does not affect this since arguments are the only thing saved here.
 
     Examples
     --------
@@ -1438,7 +1446,11 @@ def debug(func=None, prefix='', arguments=True):
     """
     if not func:
         if prefix: prefix += ' '
-        return partial(debug, prefix=prefix, arguments=arguments)
+        if out_path:
+            assert str(out_path).endswith('.json'), \
+                'out_path must ends with .json'
+        return partial(debug, prefix=prefix, arguments=arguments,
+                       out_path=out_path)
 
     @wraps(func)
     def wrapper(*args, **kwargs):
@@ -1448,13 +1460,15 @@ def debug(func=None, prefix='', arguments=True):
             sig = bound_args(wrapper, args, kwargs, collapse_kwargs=True)
             if sig:
                 first_key = next(iter(sig))
-                # Remove self/cls arg from methods.
+                # Remove self/cls arg from methods. Just check first arg to be
+                # extra careful.
                 if first_key in ('self', 'cls'):
                     del sig[first_key]
             arg_strs = (f'{k}={repr(v)}' for k, v in sig.items())
 
         # Print call message and return output.
         print(out_fmt.format(prefix, func.__qualname__, ', '.join(arg_strs)))
+        if out_path: save(dict(sig), str(out_path).format(**sig))
         return func(*args, **kwargs)
 
     return wrapper
@@ -1521,7 +1535,7 @@ def return_stdout(func):
     return wrapper
 
 
-def log_cmd(path, mode='w'):
+def log_cmd(path, mode='w', defaults=False):
     """Decorator that saves the calling command for a python script. This is
     often useful for CLIs that train ML models. It makes it easy to re-run
     the script at a later date with the same or similar arguments.
@@ -1534,6 +1548,13 @@ def log_cmd(path, mode='w'):
         Determines whether output should overwrite old file or be appended.
         One of ('a', 'w'). In most cases we will want append mode because we're
         tracking multiple trials.
+    defaults: bool
+        If True, include all arg values, even those that weren't specified
+        from the command line (e.g. if your CLI function accepts up to 10 args
+        (some with default values) and you pass in 3, the command will be
+        logged as if you explicitly passed in all 10. This can be useful if
+        you think your default args might change over time). If False, only
+        args that were explicitly mentioned in your command will be used.
 
     Examples
     --------
@@ -1573,17 +1594,32 @@ def log_cmd(path, mode='w'):
             # but here we're specifically looking at command line args. Without
             # this, I got some undesirable behavior when writing a script that
             # called another: the new command was overwriting the old log file.
-            if func.__module__ == '__main__':
-                fn_locals = bound_args(func, args, kwargs, True)
-                res = 'python'
+            if func.__module__ != '__main__': return func(*args, **kwargs)
+
+            # Log command before running script. Don't want to risk some
+            # obscure bug occurring at the end and ruining a long process.
+            fn_locals = bound_args(func, args, kwargs, True)
+            start_of_line = ' \\\n\t'
+            res = 'python'
+
+            if defaults:
+                res += __file__
+                for k, v in fn_locals.items():
+                    # Enclose data structure args in single quotes and use
+                    # double quotes inside if necessary.
+                    if isinstance(v, (tuple, list, dict, set)):
+                        v = "'" + str(v).replace("'", '"') + "'"
+                    res += f'{start_of_line}--{k} {v}'
+            else:
                 for arg in sys.argv:
-                    res += ' \\\n\t' if arg.startswith('-') else ' '
+                    res += start_of_line if arg.startswith('-') else ' '
                     # Ensure non-primitive kwargs are quoted appropriately.
                     for start, end in ['[]', '()', '{}']:
                         if arg.startswith(start) and arg.endswith(end):
                             arg = f"'{arg}'"
                     res += arg
-                save(res+'\n\n', Path(path.format(**fn_locals)), mode)
+
+            save(res+'\n\n', Path(path.format(**fn_locals)), mode)
             return func(*args, **kwargs)
         return wrapped
     return decorator
@@ -1863,3 +1899,122 @@ def immutify_defaults(func):
         return res
     return wrapper
 
+
+@contextmanager
+def temporary_global_scope(kwargs):
+    """Make a dict of key-value pairs temporarily available as global vars.
+    Used (at least as of 12/3/20) in `add_kwargs` and `fallback` decorators.
+    The original global variables should be restored after exiting the block.
+    """
+    old_globals = globals().copy()
+    globals().update(kwargs)
+    try:
+        yield
+    finally:
+        for k in kwargs:
+            if k in old_globals:
+                globals()[k] = old_globals[k]
+            else:
+                del globals()[k]
+
+
+def fallback(meth=None, *, keep=(), drop=(), save=False):
+    """Make instance/class attributes available as default arguments for a
+    method. Kwargs can be passed in to override one or more of them. You can
+    also choose for kwargs to update the instance attributes if desired.
+
+    When using default values for keep/drop/save, the decorator can be used
+    without parentheses. If you want to change one or more arguments, they
+    must be passed in as keyword args (meth is never explicitly passed in, of
+    course).
+
+    Parameters
+    ----------
+    meth: method
+        The method to decorate. Unlike the other arguments, this is passed in
+        implicitly.
+    keep: Iterable[str] or str
+        Name(s) of instance attributes to include. If you specify a value
+        here, ONLY these instance attributes will be made available as
+        fallbacks. If you don't pass in any value, the default is for all
+        instance attributes to be made available. You can specify `keep`,
+        `drop`, or neither, but not both. This covers all possible options:
+        keep only a few, keep all BUT a few, or keep all (drop all is the
+        default case and doesn't require a decorator).
+    drop: Iterable[str] or str
+        Name(s) of instance attributes to ignore. I.e. if you want to make
+        all instance attributes available as fallbacks except for self.df,
+        you could specify drop=('df').
+    save: bool
+        If True, kwargs that share names with instance attributes will be
+        overwritten with their new values. E.g. if we previously had
+        self.lr = 3e-3 and you call your decorated method with
+        obj.mymethod(lr=1), self.lr will be set to 1.
+
+    Examples
+    --------
+    # Ex 1. self.a, self.b, and self.c are all available as defaults
+
+    class Tree:
+        def __init__(self, a, b, c=3):
+            self.a = a
+            self.b = b
+            self.c = c
+
+        @fallback
+        def call(self, **kwargs):
+            return a, b, c
+
+    # Ex 2. self.b is not available as a default. We must put b in `call`'s
+    # signature or the variable won't be accessible.
+
+    class Tree:
+        def __init__(self, a, b, c=3):
+            self.a = a
+            self.b = b
+            self.c = c
+
+        @fallback(drop=('b'))
+        def call(self, b, **kwargs):
+            return a, b, c
+
+    # Ex 3. Self.b and self.c are available as defaults. If b or c are
+    # specified in kwargs, the corresponding instance attribute will be updated
+    # to take on the new value.
+
+    class Tree:
+        def __init__(self, a, b, c=3):
+            self.a = a
+            self.b = b
+            self.c = c
+
+        @fallback(keep=['b', 'c'], save=True)
+        def call(self, a, **kwargs):
+            return a, b, c
+    """
+    if meth is None:
+        # Want to avoid errors if user passes in string or leaves comma out of
+        # tuple when specifying keep/drop.
+        return partial(fallback, keep=tolist(keep), drop=tolist(drop),
+                       save=save)
+
+    @wraps(meth)
+    def wrapper(*args, **kwargs):
+        self = args[0]
+        self_kwargs = vars(self)
+        if keep or drop: self_kwargs = select(self_kwargs, keep, drop)
+
+        # Update kwargs with instance attribute defaults. Also update self if
+        # user asked to save kwargs.
+        for k, v in self_kwargs.items():
+            if k not in kwargs:
+                kwargs[k] = v
+            elif save:
+                setattr(self, k, kwargs[k])
+
+        # Execute and return.
+        with temporary_global_scope(kwargs):
+            return meth(*args, **kwargs)
+
+    return wrapper
+
